#!/usr/bin/env python

import tzupdate
import httpretty
import os
import errno
import mock
import pytest
from tests._test_utils import (
    IP_ADDRESSES,
    FAKE_SERVICES,
    FAKE_TIMEZONE,
    FAKE_ZONEINFO_PATH,
    setup_basic_api_response,
)
from hypothesis import given, settings, HealthCheck, assume
from hypothesis.strategies import sampled_from, none, one_of, text, integers


ERROR_STATUSES = [s for s in httpretty.http.STATUSES if 400 <= s <= 599]
SUPPRESSED_CHECKS = [HealthCheck.too_slow]

settings.register_profile("base", settings(suppress_health_check=SUPPRESSED_CHECKS))
settings.register_profile(
    "release", settings(max_examples=1000, suppress_health_check=SUPPRESSED_CHECKS)
)
settings.load_profile(os.getenv("HYPOTHESIS_PROFILE", "base"))


@httpretty.activate
@given(one_of(IP_ADDRESSES, none()), sampled_from(FAKE_SERVICES))
def test_get_timezone_for_ip(ip, service):
    fake_queue = mock.Mock()
    setup_basic_api_response()
    tzupdate.get_timezone_for_ip(ip_addr=ip, service=service, queue_obj=fake_queue)

    if ip is not None:
        assert ip in httpretty.last_request().path

    fake_queue.put.assert_called_once_with(FAKE_TIMEZONE)


def test_get_sys_timezone():
    systz = tzupdate.get_sys_timezone(
        FAKE_ZONEINFO_PATH, FAKE_ZONEINFO_PATH + "/" + FAKE_TIMEZONE
    )
    assert systz == FAKE_TIMEZONE


@httpretty.activate
@given(one_of(IP_ADDRESSES, none()), sampled_from(FAKE_SERVICES))
def test_get_timezone_for_ip_empty_resp(ip, service):
    fake_queue = mock.Mock()
    setup_basic_api_response(empty_resp=True)
    assert (
        tzupdate.get_timezone_for_ip(ip_addr=ip, service=service, queue_obj=fake_queue)
        is None
    )


@httpretty.activate
@given(one_of(IP_ADDRESSES, none()), sampled_from(FAKE_SERVICES))
def test_get_timezone_for_ip_empty_val(ip, service):
    fake_queue = mock.Mock()
    setup_basic_api_response(empty_val=True)
    assert (
        tzupdate.get_timezone_for_ip(ip_addr=ip, service=service, queue_obj=fake_queue)
        is None
    )


@httpretty.activate
@given(
    one_of(IP_ADDRESSES, none()),
    sampled_from(FAKE_SERVICES),
    sampled_from(ERROR_STATUSES),
)
def test_get_timezone_for_ip_doesnt_raise(ip, service, status):
    fake_queue = mock.Mock()
    setup_basic_api_response(status=status)
    assert (
        tzupdate.get_timezone_for_ip(ip_addr=ip, service=service, queue_obj=fake_queue)
        is None
    )


@mock.patch("tzupdate.os.unlink")
@mock.patch("tzupdate.os.symlink")
@mock.patch("tzupdate.os.path.isfile")
def test_link_localtime(isfile_mock, symlink_mock, unlink_mock):
    isfile_mock.return_value = True
    expected = os.path.join(tzupdate.DEFAULT_ZONEINFO_PATH, FAKE_TIMEZONE)

    zoneinfo_tz_path = tzupdate.link_localtime(
        FAKE_TIMEZONE, tzupdate.DEFAULT_ZONEINFO_PATH, tzupdate.DEFAULT_LOCALTIME_PATH
    )

    assert unlink_mock.called_once_with([expected])
    assert symlink_mock.called_once_with([expected, tzupdate.DEFAULT_LOCALTIME_PATH])

    assert zoneinfo_tz_path == expected


@given(text())
def test_link_localtime_traversal_attack_root(questionable_timezone):
    assume(tzupdate.DEFAULT_ZONEINFO_PATH not in questionable_timezone)
    questionable_timezone = "/" + questionable_timezone

    with pytest.raises(tzupdate.DirectoryTraversalError):
        tzupdate.link_localtime(
            questionable_timezone,
            tzupdate.DEFAULT_ZONEINFO_PATH,
            tzupdate.DEFAULT_LOCALTIME_PATH,
        )


@given(text())
def test_link_localtime_traversal_attack_dotdot(questionable_timezone):
    assume(tzupdate.DEFAULT_ZONEINFO_PATH not in questionable_timezone)
    questionable_timezone = "../../../" + questionable_timezone

    with pytest.raises(tzupdate.DirectoryTraversalError):
        tzupdate.link_localtime(
            questionable_timezone,
            tzupdate.DEFAULT_ZONEINFO_PATH,
            tzupdate.DEFAULT_LOCALTIME_PATH,
        )


@mock.patch("tzupdate.os.path.isfile")
def test_link_localtime_timezone_not_available(isfile_mock):
    isfile_mock.return_value = False
    with pytest.raises(tzupdate.TimezoneNotLocallyAvailableError):
        tzupdate.link_localtime(
            FAKE_TIMEZONE,
            tzupdate.DEFAULT_ZONEINFO_PATH,
            tzupdate.DEFAULT_LOCALTIME_PATH,
        )


@mock.patch("tzupdate.os.unlink")
@mock.patch("tzupdate.os.path.isfile")
def test_link_localtime_permission_denied(isfile_mock, unlink_mock):
    isfile_mock.return_value = True
    unlink_mock.side_effect = OSError(errno.EACCES, "Permission denied yo")
    with pytest.raises(OSError) as raise_cm:
        tzupdate.link_localtime(
            FAKE_TIMEZONE,
            tzupdate.DEFAULT_ZONEINFO_PATH,
            tzupdate.DEFAULT_LOCALTIME_PATH,
        )

    assert raise_cm.value.errno == errno.EACCES


@mock.patch("tzupdate.os.unlink")
@mock.patch("tzupdate.os.path.isfile")
def test_link_localtime_oserror_not_permission(isfile_mock, unlink_mock):
    isfile_mock.return_value = True
    code = errno.ENOSPC
    unlink_mock.side_effect = OSError(code, "No space yo")

    with pytest.raises(OSError) as thrown_exc:
        tzupdate.link_localtime(
            FAKE_TIMEZONE,
            tzupdate.DEFAULT_ZONEINFO_PATH,
            tzupdate.DEFAULT_LOCALTIME_PATH,
        )

    assert thrown_exc.value.errno == code


@mock.patch("tzupdate.os.unlink")
@mock.patch("tzupdate.os.path.isfile")
@mock.patch("tzupdate.os.symlink")
def test_link_localtime_localtime_missing_no_raise(
    symlink_mock, isfile_mock, unlink_mock
):
    isfile_mock.return_value = True
    code = errno.ENOENT
    unlink_mock.side_effect = OSError(code, "No such file or directory")

    # This should handle OSError and not raise further
    tzupdate.link_localtime(
        FAKE_TIMEZONE, tzupdate.DEFAULT_ZONEINFO_PATH, tzupdate.DEFAULT_LOCALTIME_PATH
    )


@given(text(), text())
def test_debian_tz_path_exists_not_forced(timezone, tz_path):
    mo = mock.mock_open()
    with mock.patch("tzupdate.open", mo, create=True):
        tzupdate.write_debian_timezone(timezone, tz_path, must_exist=True)
        mo.assert_called_once_with(tz_path, "r+")
        mo().seek.assert_called_once_with(0)
        mo().write.assert_called_once_with(timezone + "\n")


@given(text(), text())
def test_debian_tz_path_doesnt_exist_not_forced(timezone, tz_path):
    mo = mock.mock_open()
    mo.side_effect = OSError(errno.ENOENT, "")
    with mock.patch("tzupdate.open", mo, create=True):
        tzupdate.write_debian_timezone(timezone, tz_path, must_exist=True)
        mo.assert_called_once_with(tz_path, "r+")


@given(text(), text())
def test_debian_tz_path_other_error_raises(timezone, tz_path):
    mo = mock.mock_open()
    code = errno.EPERM
    mo.side_effect = OSError(code, "")
    with mock.patch("tzupdate.open", mo, create=True):
        with pytest.raises(OSError) as thrown_exc:
            tzupdate.write_debian_timezone(timezone, tz_path, must_exist=True)
        assert thrown_exc.value.errno == code


@given(text(), text())
def test_debian_tz_path_doesnt_exist_forced(timezone, tz_path):
    mo = mock.mock_open()
    with mock.patch("tzupdate.open", mo, create=True):
        tzupdate.write_debian_timezone(timezone, tz_path, must_exist=False)
        mo.assert_called_once_with(tz_path, "w")
        mo().seek.assert_called_once_with(0)
        mo().write.assert_called_once_with(timezone + "\n")



#!/usr/bin/env python

import httpretty
import tzupdate
import mock
import pytest
from tests._test_utils import FAKE_SERVICES, FAKE_TIMEZONE, setup_basic_api_response


@httpretty.activate
@mock.patch("tzupdate.write_debian_timezone")
@mock.patch("tzupdate.link_localtime")
def test_end_to_end_no_args(link_localtime_mock, deb_tz_mock):
    setup_basic_api_response()
    args = []
    tzupdate.main(args, services=FAKE_SERVICES)
    link_localtime_mock.assert_called_once_with(
        FAKE_TIMEZONE, tzupdate.DEFAULT_ZONEINFO_PATH, tzupdate.DEFAULT_LOCALTIME_PATH
    )
    deb_tz_mock.assert_called_once_with(
        FAKE_TIMEZONE, tzupdate.DEFAULT_DEBIAN_TIMEZONE_PATH, True
    )


@httpretty.activate
@mock.patch("tzupdate.write_debian_timezone")
@mock.patch("tzupdate.link_localtime")
def test_print_only_no_link(link_localtime_mock, deb_tz_mock):
    setup_basic_api_response()
    args = ["-p"]
    tzupdate.main(args, services=FAKE_SERVICES)
    assert not link_localtime_mock.called
    assert not deb_tz_mock.called


@mock.patch("tzupdate.write_debian_timezone")
@mock.patch("tzupdate.link_localtime")
def test_print_sys_tz_no_link(link_localtime_mock, deb_tz_mock):
    args = ["--print-system-timezone"]
    tzupdate.main(args, services=FAKE_SERVICES)
    assert not link_localtime_mock.called
    assert not deb_tz_mock.called


@httpretty.activate
@mock.patch("tzupdate.write_debian_timezone")
@mock.patch("tzupdate.link_localtime")
def test_explicit_paths(link_localtime_mock, deb_tz_mock):
    setup_basic_api_response()
    localtime_path = "/l"
    zoneinfo_path = "/z"
    deb_path = "/d"
    args = ["-l", localtime_path, "-z", zoneinfo_path, "-d", deb_path]
    tzupdate.main(args, services=FAKE_SERVICES)
    link_localtime_mock.assert_called_once_with(
        FAKE_TIMEZONE, zoneinfo_path, localtime_path
    )
    deb_tz_mock.assert_called_once_with(FAKE_TIMEZONE, deb_path, True)


@mock.patch("tzupdate.write_debian_timezone")
@mock.patch("tzupdate.link_localtime")
@mock.patch("tzupdate.get_timezone")
def test_explicit_ip(get_timezone_mock, _unused_ll, _unused_deb):
    ip_addr = "1.2.3.4"
    args = ["-a", ip_addr]
    tzupdate.main(args, services=FAKE_SERVICES)
    get_timezone_mock.assert_called_once_with(
        ip_addr, timeout=mock.ANY, services=FAKE_SERVICES
    )


@mock.patch("tzupdate.write_debian_timezone")
@mock.patch("tzupdate.link_localtime")
def test_explicit_timezone(link_localtime_mock, deb_tz_mock):
    timezone = "Foo/Bar"
    args = ["-t", timezone]
    tzupdate.main(args)
    link_localtime_mock.assert_called_once_with(
        timezone, tzupdate.DEFAULT_ZONEINFO_PATH, tzupdate.DEFAULT_LOCALTIME_PATH
    )
    deb_tz_mock.assert_called_once_with(
        timezone, tzupdate.DEFAULT_DEBIAN_TIMEZONE_PATH, True
    )


@httpretty.activate
@mock.patch("tzupdate.Process")
def test_timeout_results_in_exception(process_mock):
    # The process mock causes us to never run get_timezone_from_ip, so we
    # should time out
    setup_basic_api_response()
    args = ["-s", "0.01"]
    with pytest.raises(tzupdate.TimezoneAcquisitionError):
        tzupdate.main(args, services=FAKE_SERVICES)



#!/usr/bin/env python

import tzupdate
import httpretty
import json
import re
import ipaddress
from hypothesis.strategies import integers, builds

try:
    _ipa_type = unicode  # pytype: disable=name-error
except NameError:
    _ipa_type = str

IP_ADDRESSES = builds(
    ipaddress.IPv4Address, integers(min_value=0, max_value=(2 ** 32 - 1))
).map(_ipa_type)

FAKE_TIMEZONE = "Fake/Timezone"
FAKE_ZONEINFO_PATH = "/path/to/zoneinfo"
FAKE_ERROR = "Virus = very yes"

FAKE_SERVICES = [
    tzupdate.GeoIPService("http://example.com/json/{ip}", ("timezone",), ("message",)),
    tzupdate.GeoIPService("https://doesnotexistreally.com/{ip}", ("time_zone",), None),
    tzupdate.GeoIPService(
        "http://tzupdate.com/foo/bar/{ip}", ("location", "time_zone"), ("msg",)
    ),
]


def setup_basic_api_response(
    services=None, empty_resp=False, empty_val=False, status=200
):
    """
    If `empty_resp', we return a totally empty API response, except for any
    error message.

    If `empty_val', the tz_key is an empty string.
    """
    if services is None:
        services = FAKE_SERVICES

    for service in services:
        url_regex = re.compile(service.url.format(ip=r".*"))
        api_body = {}

        if not empty_resp:
            cur_level = api_body

            for i, key in enumerate(service.tz_keys, start=1):
                if i == len(service.tz_keys):
                    if empty_val:
                        cur_level[key] = ""
                    else:
                        cur_level[key] = FAKE_TIMEZONE
                else:
                    cur_level[key] = {}
                    cur_level = cur_level[key]

        tmp = {}
        if service.error_keys:
            for i, key in enumerate(reversed(service.error_keys)):
                if i == 0:
                    tmp[key] = FAKE_ERROR
                else:
                    tmp[key] = tmp.copy()

        httpretty.register_uri(
            httpretty.GET,
            url_regex,
            body=json.dumps(api_body),
            content_type="application/json",
            status=status,
        )



#!/usr/bin/env python

"""
Set the system timezone based on IP geolocation.
"""

from __future__ import print_function

from multiprocessing import Queue, Process
import argparse
import collections
import errno
import logging
import json
import os
import sys

from queue import Empty
from urllib.request import urlopen
from urllib.error import HTTPError

LOG = logging.getLogger(__name__)

DEFAULT_ZONEINFO_PATH = "/usr/share/zoneinfo"
DEFAULT_LOCALTIME_PATH = "/etc/localtime"
DEFAULT_DEBIAN_TIMEZONE_PATH = "/etc/timezone"
DEFAULT_HTTP_TIMEOUT = 30.0


# url: A url with an "ip" key to be replaced with an optional IP
# tz_keys: The key hierarchy to get the timezone from
# error_keys: Optionally, where to get error messages from
GeoIPService = collections.namedtuple("GeoIPService", ["url", "tz_keys", "error_keys"])

SERVICES = frozenset(
    [
        GeoIPService("http://ip-api.com/json/{ip}", ("timezone",), ("message",)),
        GeoIPService("https://freegeoip.app/json/{ip}", ("time_zone",), None),
        GeoIPService("https://ipapi.co/{ip}/json/", ("timezone",), ("reason",)),
        GeoIPService("http://worldtimeapi.org/api/ip/{ip}", ("timezone",), ("error",)),
    ]
)


def get_deep(item, keys):
    tmp = item
    for key in keys:
        tmp = tmp[key]
    return tmp


def get_timezone(ip_addr, timeout=DEFAULT_HTTP_TIMEOUT, services=SERVICES):
    api_resp_queue = Queue()

    threads = [
        Process(target=get_timezone_for_ip, args=(ip_addr, svc, api_resp_queue))
        for svc in services
    ]

    for thread in threads:
        thread.start()

    try:
        timezone = api_resp_queue.get(block=True, timeout=timeout)
    except Empty:
        raise TimezoneAcquisitionError(
            "No usable response from any API in {} seconds. Consider "
            "increasing --timeout if your connection is slow.".format(timeout)
        ) from None
    finally:
        for thread in threads:
            thread.terminate()

    return timezone


def get_timezone_for_ip(ip_addr, service, queue_obj):
    api_url = service.url.format(ip=ip_addr or "")

    try:
        # The caller is responsible for providing a service string which
        # doesn't permit walking file: URIs or whatever, so silence bandit's
        # warning about that
        api_response_obj = urlopen(api_url)  # nosec
    except HTTPError as thrown_exc:
        LOG.warning("%s returned %d, ignoring", api_url, thrown_exc.code)
        return

    api_response = json.loads(api_response_obj.read().decode("utf8"))
    LOG.debug("API response from %s: %r", api_url, api_response)

    try:
        timezone = get_deep(api_response, service.tz_keys)
        if not timezone:
            raise KeyError
    except KeyError:
        msg = "Unspecified API error for {}.".format(service.url)
        if service.error_keys is not None:
            try:
                msg = get_deep(api_response, service.error_keys)
            except KeyError:
                pass
        LOG.warning("%s failed: %s", api_url, msg)
    else:
        queue_obj.put(timezone)


def write_debian_timezone(timezone, debian_timezone_path, must_exist=True):
    """
    Debian and derivatives also have /etc/timezone, which is used for a human
    readable timezone. Without this, dpkg-reconfigure will nuke /etc/localtime
    on reconfigure.

    If must_exist is True, we won't create debian_timezone_path if it doesn't
    already exist.
    """
    old_umask = os.umask(0o133)
    mode = "w"

    if must_exist:
        mode = "r+"

    try:
        with open(debian_timezone_path, mode) as debian_tz_f:
            debian_tz_f.seek(0)
            debian_tz_f.write(timezone + "\n")
    except OSError as thrown_exc:
        if must_exist and thrown_exc.errno == errno.ENOENT:
            return
        raise
    finally:
        os.umask(old_umask)


def check_directory_traversal(base_dir, requested_path):
    """
    Check for directory traversal, and raise an exception if it was detected.

    Since we are linking based upon the output of some data we retrieved over
    the internet, we should check that it doesn't attempt to do something
    naughty with local path traversal.

    This function checks that the base directory of the zoneinfo database
    shares a common prefix with the absolute path of the requested zoneinfo
    file.
    """
    LOG.debug("Checking for traversal in path %s", requested_path)
    requested_path_abs = os.path.abspath(requested_path)
    LOG.debug("Absolute path of requested path is %s", requested_path_abs)
    if os.path.commonprefix([base_dir, requested_path_abs]) != base_dir:
        raise DirectoryTraversalError(
            "%r (%r) is outside base directory %r, refusing to run"
            % (requested_path, requested_path_abs, base_dir)
        )


def link_localtime(timezone, zoneinfo_path, localtime_path):
    """
    Link a timezone file from the zoneinfo database to /etc/localtime.

    Since we may be retrieving the timezone file's relative path from an
    untrusted source, we also do checks to make sure that no directory
    traversal is going on. See `check_directory_traversal` for information
    about how that works.
    """
    zoneinfo_tz_path = os.path.join(zoneinfo_path, timezone)
    check_directory_traversal(zoneinfo_path, zoneinfo_tz_path)

    if not os.path.isfile(zoneinfo_tz_path):
        raise TimezoneNotLocallyAvailableError(
            'Timezone "%s" requested, but this timezone is not available on '
            "your operating system." % timezone
        )

    try:
        os.unlink(localtime_path)
    except OSError as thrown_exc:
        # If we don't have permission to unlink /etc/localtime, we probably
        # need to be root.
        if thrown_exc.errno == errno.EACCES:
            raise OSError(
                thrown_exc.errno,
                'Could not link "%s" (%s). Are you root?'
                % (localtime_path, thrown_exc),
            )
        if thrown_exc.errno != errno.ENOENT:
            raise

    os.symlink(zoneinfo_tz_path, localtime_path)

    return zoneinfo_tz_path


def get_sys_timezone(zoneinfo_abspath, localtime_abspath):
    return localtime_abspath.replace(
        os.path.commonprefix([zoneinfo_abspath, localtime_abspath]) + os.path.sep, "", 1
    )


def parse_args(argv):
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument(
        "-p",
        "--print-only",
        action="store_true",
        help="print the timezone, but don't update the localtime file",
    )
    parser.add_argument(
        "--print-system-timezone",
        action="store_true",
        help="print the current system timezone",
    )
    parser.add_argument(
        "-a", "--ip", help="use this IP instead of automatically detecting it"
    )
    parser.add_argument(
        "-t",
        "--timezone",
        help="use this timezone instead of automatically detecting it",
    )
    parser.add_argument(
        "-z",
        "--zoneinfo-path",
        default=DEFAULT_ZONEINFO_PATH,
        help="path to root of the zoneinfo database (default: %(default)s)",
    )
    parser.add_argument(
        "-l",
        "--localtime-path",
        default=DEFAULT_LOCALTIME_PATH,
        help="path to localtime symlink (default: %(default)s)",
    )
    parser.add_argument(
        "-d",
        "--debian-timezone-path",
        default=DEFAULT_DEBIAN_TIMEZONE_PATH,
        help="path to Debian timezone name file (default: %(default)s)",
    )
    parser.add_argument(
        "--always-write-debian-timezone",
        action="store_true",
        help="create debian timezone file even if it doesn't exist (default: %(default)s)",
    )
    parser.add_argument(
        "-s",
        "--timeout",
        help="maximum number of seconds to wait for APIs to return (default: "
        "%(default)s)",
        type=float,
        default=DEFAULT_HTTP_TIMEOUT,
    )
    parser.add_argument(
        "--debug",
        action="store_const",
        dest="log_level",
        const=logging.DEBUG,
        default=logging.WARNING,
        help="enable debug logging",
    )
    args = parser.parse_args(argv)
    return args


def main(argv=None, services=SERVICES):
    if argv is None:  # pragma: no cover
        argv = sys.argv[1:]

    args = parse_args(argv)
    logging.basicConfig(level=args.log_level)

    if args.print_system_timezone:
        print(
            get_sys_timezone(
                os.path.realpath(args.zoneinfo_path),
                os.path.realpath(args.localtime_path),
            )
        )
        return

    if args.timezone:
        timezone = args.timezone
        LOG.debug("Using explicitly passed timezone: %s", timezone)
    else:
        timezone = get_timezone(args.ip, timeout=args.timeout, services=services)

    if args.print_only:
        print(timezone)
    else:
        link_localtime(timezone, args.zoneinfo_path, args.localtime_path)
        write_debian_timezone(
            timezone, args.debian_timezone_path, not args.always_write_debian_timezone
        )
        print("Set system timezone to %s." % timezone)


class TimezoneUpdateException(Exception):
    """
    Base class for exceptions raised by tzupdate.
    """


class TimezoneNotLocallyAvailableError(TimezoneUpdateException):
    """
    Raised when the API returned a timezone, but we don't have it locally.
    """


class DirectoryTraversalError(TimezoneUpdateException):
    """
    Raised when the timezone path returned by the API would result in directory
    traversal when concatenated with the zoneinfo path.
    """


class TimezoneAcquisitionError(TimezoneUpdateException):
    """
    Raised when all timezone APIs do not return in a timely manner.
    """


if __name__ == "__main__":  # pragma: no cover
    main()



#!/usr/bin/env python

from setuptools import setup


with open("README.rst") as readme_f:
    README = readme_f.read()

setup(
    name="tzupdate",
    version="2.0.0",
    description="Set the system timezone based on IP geolocation",
    long_description=README,
    url="https://github.com/cdown/tzupdate",
    license="Public Domain",
    author="Chris Down",
    author_email="chris@chrisdown.name",
    py_modules=["tzupdate"],
    entry_points={"console_scripts": ["tzupdate=tzupdate:main"]},
    keywords="timezone localtime tz",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "License :: Public Domain",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Topic :: System",
        "Topic :: System :: Networking :: Time Synchronization",
        "Topic :: Utilities",
    ],
)
